/* mul_2nm1.c
 *
 * Experimental code for multiplication mod 2^n-1.

Copyright 2009, Niels Mller.

This file is not yet part of anything, but intended for GMP.

This file is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

This file is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this file.  If not, see http://www.gnu.org/licenses/.  */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

/* Manually tuned. */
#if 0
/* for x86_32 (AMD K7) */
static mp_size_t x2_threshold = 22;
static mp_size_t dc_x2_threshold = 44;
#elif 1
/* for p6 */
static mp_size_t x2_threshold = 16;
static mp_size_t dc_threshold = 16;
#else
/* Minimal */
static mp_size_t x2_threshold = 2;
static mp_size_t dc_threshold = 4;
#endif

/* Inputs and outputs are n limbs, computation is mod B^n - 1, and
 * values are seminormalized; zero is represented as either 0 or B^n -
 * 1. Needs 2n limbs at rp. */
static void
mul_2nm1_wrap (mp_ptr rp, mp_srcptr ap, mp_srcptr bp, mp_size_t n)
{
  mp_limb_t cy;
  mpn_mul_n (rp, ap, bp, n);
  cy = mpn_add_n (rp, rp, rp+n, n);
  /* If cy == 1, then the value of rp is at most B^n - 2, so there can
   * be no overflow when adding in the carry. */
  MPN_INCR_U (rp, n, cy);
}

#define MUL_2NM1_X2_ITCH(n) (4*(n)+3)

/* Evaluation in 1 and -1. Needs size+1 limbs at rp. */
static __attribute__((noinline)) void
mul_2nm1_x2 (mp_ptr rp, mp_srcptr ap, mp_srcptr bp, mp_size_t size,
	     mp_ptr tp)
{
  mp_limb_t n;
  mp_limb_t cy;
  int neg;

  /* For now, support only even n */
  ASSERT (! (size & 1));
  n = size / 2;
#define a0 ap
#define a1 (ap + n)
#define b0 bp
#define b1 (bp + n)

#define ap1 tp			/* n + 1 */
#define am1 (tp + n + 1)	/* n */
#define bp1 (tp + 2*n + 1)	/* n + 1 */
#define bm1 (tp + 3*n + 2)	/* n */
#define vp1 (tp + 4*n + 2)	/* 2n + 1 (+ 1) */
#define vm1 (tp + 6*n + 3)	/* 2n */

  /* Total scratch need is 8*n + 3 = 4*size + 3 */
  /* Compute a0 + a1 and a0 - a1 */
  ap1[n] = mpn_add_n (ap1, a0, a1, n);
  if (mpn_cmp (a0, a1, n) < 0)
    {
      ASSERT_NOCARRY (mpn_sub_n (am1, a1, a0, n));
      neg = 1;
    }
  else
    {
      ASSERT_NOCARRY (mpn_sub_n (am1, a0, a1, n));
      neg = 0;
    }

  /* Compute b0 + b1 and b0 - b1 */
  bp1[n] = mpn_add_n (bp1, b0, b1, n);
  if (mpn_cmp (b0, b1, n) < 0)
    {
      ASSERT_NOCARRY (mpn_sub_n (bm1, b1, b0, n));
      neg ^= 1;
    }
  else
    {
      ASSERT_NOCARRY (mpn_sub_n (bm1, b0, b1, n));
    }

  mpn_mul_n (vp1, ap1, bp1, n+1);
  ASSERT (vp1[2*n + 1] == 0);
  ASSERT (vp1[2*n] < 4);

  mpn_mul_n (vm1, am1, bm1, n);
  
  if (neg)
    ASSERT_NOCARRY (mpn_sub (rp, vp1, 2*n+1, vm1, 2*n));
  else
    ASSERT_NOCARRY (mpn_add (rp, vp1, 2*n+1, vm1, 2*n));

  ASSERT_NOCARRY (mpn_rshift (rp, rp, 2*n+1, 1));

  ASSERT_NOCARRY (mpn_sub_n (vp1, vp1, rp, 2*n+1));
  
  ASSERT_NOCARRY (mpn_add (rp, rp, 2*n+1, vp1 + n, n+1));
  ASSERT_NOCARRY (mpn_add (rp + n, rp + n, n+1, vp1, n));
  cy = mpn_add_1 (rp, rp, 2*n, rp[2*n]);

  MPN_INCR_U (rp, 2*n, cy);

#undef a0
#undef a1
#undef b0
#undef b1
#undef ap1
#undef am1
#undef bp1
#undef bm1
#undef vp1
#undef vm1
}

enum eval_dgr4_flags
  {
    eval_dgr4_m1_neg = 1,
    eval_dgr4_re_neg = 2,
    eval_dgr4_im_neg = 4,
    eval_dgr4_sum_neg = 8,
  };
    
static enum eval_dgr4_flags
eval_dgr4 (mp_ptr p1, mp_ptr m1, mp_ptr re, mp_ptr im, mp_ptr sum,
	   mp_srcptr ap, mp_size_t n)
{
  /* Use the points
   *
   * ap1 = a(1)     = a0 + a1 + a2 + a3
   * am1 = a(-1)    = a0 - a1 + a2 - a3
   * are = Re(a(i)) = a0      - a2
   * aim = Im(a(i)) =      a1      - a3
   *
   * For the complex multiplication, we also need
   *
   * asum = are+aim = a0 + a1 - a2 - a3 
   */
#define a0 ap
#define a1 (ap + n)
#define a2 (ap + 2*n)
#define a3 (ap + 3*n)

  enum eval_dgr4_flags flags;

  /* FIXME: This function can make a lot of use of add_n_sub_n. */  
  
  /* Evaluate in 1 and -1 */
  p1[n] = mpn_add_n (p1, a0, a2, n);
  re[n] = mpn_add_n (re, a1, a3, n);
  if (mpn_cmp (p1, re, n+1) < 0)
    {
      ASSERT_NOCARRY (mpn_sub_n (m1, re, p1, n+1));
      flags = eval_dgr4_m1_neg;
    }
  else
    {
      ASSERT_NOCARRY (mpn_sub_n (m1, p1, re, n+1));
      flags = 0;
    }
  ASSERT_NOCARRY (mpn_add_n (p1, p1, re, n+1));

  /* Real part of value at i */
  if (mpn_cmp (a0, a2, n) < 0)
    {
      ASSERT_NOCARRY (mpn_sub_n (re, a2, a0, n));
      flags |= eval_dgr4_re_neg;
    }
  else
    ASSERT_NOCARRY (mpn_sub_n (re, a0, a2, n));

  /* Imaginary part of value at i */
  if (mpn_cmp (a1, a3, n) < 0)
    {
      ASSERT_NOCARRY (mpn_sub_n (im, a3, a1, n));
      flags |= eval_dgr4_im_neg;
    }
  else
    ASSERT_NOCARRY (mpn_sub_n (im, a1, a3, n));

  if ( (flags ^ (flags << 1)) & eval_dgr4_im_neg)
    {
      /* Real and imaginary value have different signs. */
      if (mpn_cmp (re, im, n) < 0)
	{
	  ASSERT_NOCARRY (mpn_sub_n (sum, im, re, n));
	  flags |= (flags << 1) & eval_dgr4_sum_neg;
	}
      else
	{
	  ASSERT_NOCARRY (mpn_sub_n (sum, re, im, n));
	  flags |= (flags << 2) & eval_dgr4_sum_neg;
	}
      sum[n] = 0;
    }
  else
    {
      /* Real and imaginary value have same sign. */
      sum[n] = mpn_add_n (sum, re, im, n);
      flags |= (flags << 1) & eval_dgr4_sum_neg;      
    }
  return flags;
#undef a0
#undef a1
#undef a2
#undef a3
}

#define MUL_2NM1_X4_ITCH(n) (5*(n)+11)

/* Evaluation in 1, -1 and i. Needs size+1 limbs at rp. */
static __attribute__((noinline)) void
mul_2nm1_x4 (mp_ptr rp, mp_srcptr ap, mp_srcptr bp, mp_size_t size,
	     mp_ptr tp)
{  
  mp_size_t n;
  enum eval_dgr4_flags flags;
  mp_limb_t hi;
  
  /* For now, support only sizes that are multiples of 4. */
  ASSERT (! (size & 3));
  n = size / 4;

#define ap1 tp			/* n + 1 */
#define am1 (tp + n + 1)	/* n + 1 */
#define are (tp + 2*n + 2)	/* n */
#define aim (tp + 3*n + 2)	/* n */
#define asum (tp + 4*n + 2)	/* n + 1 */
  
#define bp1 (tp + 5*n + 3)	/* n + 1 */
#define bm1 (tp + 6*n + 4)	/* n + 1 */
#define bre (tp + 7*n + 5)	/* n */
#define bim (tp + 8*n + 5)	/* n */
#define bsum (tp + 9*n + 5)	/* n + 1 */

#define vp1 (tp + 10*n + 6)	/* 2n + 1 (+ 1) */
#define vm1 (tp + 12*n + 7)	/* 2n + 1 (+ 1) */
#define vsum (tp + 14*n + 8)	/* 2n + 1 (+ 1) */
#define vre (tp + 16*n + 9)	/* 2n + 1*/
#define vim (tp + 18*n + 10)	/* 2n + 1*/

  /* Total scratch need: 20*n + 9 = 5*size + 11 */

  flags = eval_dgr4 (ap1, am1, are, aim, asum, ap, n);
  flags ^= eval_dgr4 (bp1, bm1, bre, bim, bsum, bp, n);

  mpn_mul_n (vp1, ap1, bp1, n+1);
  ASSERT (vp1[2*n+1] == 0);
  ASSERT (vp1[2*n] < 16);
  mpn_mul_n (vm1, am1, bm1, n+1);
  ASSERT (vm1[2*n+1] == 0);
  ASSERT (vm1[2*n] < 16);
  mpn_mul_n (vsum, asum, bsum, n+1);
  ASSERT (vsum[2*n+1] == 0);
  ASSERT (vsum[2*n] < 4);
  mpn_mul_n (vre, are, bre, n);
  mpn_mul_n (vim, aim, bim, n);

  /* Complex multiplication,
   *
   * (are + i aim)*(bre + i bim) = (are*bre - aim*bim) + i (asum*bsum - are*bre - aim*bim)
   *
   * Values stored at tp are no longer used, so reuse for the new real
   * part. FIXME: Could do the butterfly operation with no scratch
   * space, as x <- x + y, y <- x - 2*y. FIXME: Could also make good
   * use of mpn_add_n_sub_n. */

  /* Set {tp, 2*n+1} = vre + vim, vre = vre - vim,  */
  if ( (flags ^ (flags << 1)) & eval_dgr4_im_neg)
    {
      /* vre and vim have different signs. */
      if (mpn_cmp (vre, vim, 2*n) < 0)
	{
	  ASSERT_NOCARRY (mpn_sub_n (tp, vim, vre, 2*n));
	}
      else
	{
	  ASSERT_NOCARRY (mpn_sub_n (tp, vre, vim, 2*n));
	  flags ^= eval_dgr4_im_neg;
	}
      tp[2*n] = 0;
      vre[2*n] = mpn_add_n (vre, vim, vre, 2*n);
      
    }
  else
    {
      /* vre and vim have the same sign */
      tp[2*n] = mpn_add_n (tp, vim, vre, 2*n);

      if (mpn_cmp (vre, vim, 2*n) < 0)
	{
	  ASSERT_NOCARRY (mpn_sub_n (vre, vim, vre, 2*n));
	  flags ^= eval_dgr4_re_neg;	  
	}
      else
	{
	  ASSERT_NOCARRY (mpn_sub_n (vre, vre, vim, 2*n));
	}
      vre[2*n] = 0;
    }

  /* vim = vsum - vim */
  if ( (flags ^ (flags << 1)) & eval_dgr4_sum_neg)
    {
      /* Different signs */
      ASSERT_NOCARRY (mpn_add_n (vim, tp, vsum, 2*n+1));
      flags ^= eval_dgr4_im_neg;
    }
  else
    {
      /* Same signs */
      if (mpn_cmp (vsum, tp, 2*n+1) < 0)
	{
	  ASSERT_NOCARRY (mpn_sub_n (vim, tp, vsum, 2*n + 1));
	  flags ^= eval_dgr4_im_neg;
	}
      else
	{
	  ASSERT_NOCARRY (mpn_sub_n (vim, vsum, tp, 2*n + 1));
	}
    }

  /* vm1 = (vp1 + vm1) / 2 = x0 + x2 */
  if (flags & eval_dgr4_m1_neg)
    ASSERT_NOCARRY (mpn_sub_n (vm1, vp1, vm1, 2*n + 1));
  else
    ASSERT_NOCARRY (mpn_add_n (vm1, vp1, vm1, 2*n + 1));

  ASSERT_NOCARRY (mpn_rshift (vm1, vm1, 2*n+1, 1));

  /* vp1 = vp1 - vm1 = x1 + x3 */
  ASSERT_NOCARRY (mpn_sub_n (vp1, vp1, vm1, 2*n+1));

  /* {rp, 2*n+1} = (vm1 + vre) / 2 = x0 */
  if (flags & eval_dgr4_re_neg)
    ASSERT_NOCARRY (mpn_sub_n (rp, vm1, vre, 2*n+1));
  else
    ASSERT_NOCARRY (mpn_add_n (rp, vm1, vre, 2*n+1));

  ASSERT_NOCARRY (mpn_rshift (rp, rp, 2*n+1, 1));

  /* vm1 = vm1 - {rp, 2*n+1} = x2 */
  ASSERT_NOCARRY (mpn_sub_n (vm1, vm1, rp, 2*n+1));
  
  /* vim = (vp1 + vim) / 2 = x1 */
  if (flags & eval_dgr4_im_neg)
    ASSERT_NOCARRY (mpn_sub_n (vim, vp1, vim, 2*n+1));
  else
    ASSERT_NOCARRY (mpn_add_n (vim, vp1, vim, 2*n+1));
  ASSERT_NOCARRY (mpn_rshift (vim, vim, 2*n+1, 1));
  
  /* vp1 = vp1 - vim = x3 */
  ASSERT_NOCARRY (mpn_sub_n (vp1, vp1, vim, 2*n+1));

  /* Now, it's substitution time, we want to compute
   *
   * x0 + x1 B + x2 B^2 + x3 B^3 mod B^4, so the last coafficient
   * wraps around */

  ASSERT_NOCARRY (mpn_add (rp + n, vim, 2*n + 1, rp + n, n+1));
  hi = vm1[2*n] + mpn_add (rp + 2*n, vm1, 2*n, rp + 2*n, n+1);
  hi += mpn_add_n (rp + 3*n, rp + 3*n, vp1, n);
  MPN_INCR_U (vp1+n, n+1, hi);
  
  hi = mpn_add (rp, rp, 4*n, vp1 + n, n+1);
  MPN_INCR_U (rp, 4*n, hi);  

#undef ap1
#undef am1
#undef are
#undef aim
#undef asum
#undef bp1
#undef bm1
#undef bre
#undef bim
#undef bsum
#undef vp1
#undef vn1
#undef vsum
#undef vre
#undef vim
}

/* Inputs and outputs are n + 1, in semi-normalized representation,
   computation is mod B^n + 1. Needs 2n + 2 limbs at rp. Output is
   normalized. */
static void
mul_2np1_wrap (mp_ptr rp, mp_srcptr ap, mp_srcptr bp, mp_size_t n)
{
  mp_limb_t cy;
  mpn_mul_n (rp, ap, bp, n+1);
  cy = mpn_add (rp, rp, n, rp + 2*n, 2);
  cy = mpn_sub_nc (rp, rp, rp+n, n, cy);
  rp[n] = mpn_add_1 (rp, rp, n, cy);
}

/* Scratch need: 3*n + 4 + need for recursive call. This gives
 *
 * S(n) <= 3*n + 4 + S(n/2) <= 6n + c log n
 */
#define MUL_2NM1_DC_ITCH(n) (6*(n) + GMP_LIMB_BITS)

static __attribute__((noinline)) void
mul_2nm1_dc (mp_ptr rp, mp_srcptr ap, mp_srcptr bp, mp_size_t size,
	     mp_ptr tp)
{
  if ( (size & 1) || size < x2_threshold)
    mul_2nm1_wrap (rp, ap, bp, size);
  else if (size < dc_threshold)
    mul_2nm1_x2 (rp, ap, bp, size, tp);
  else
    {
      mp_size_t n;
      mp_limb_t cy;
      mp_limb_t hi;

      n = size / 2;
  
      /* Compute xm = a*b mod (B^n - 1), xp = a*b mod (B^n + 1)
	 and crt together as

	 x = xm + (B^n - 1) * [B^n/2 (xp - xm) mod (B^n+1)]
      */

#define a0 ap
#define a1 (ap + n)
#define b0 bp
#define b1 (bp + n)

#define am1 tp		/* n */
#define bm1 (tp + n)	/* n */
#define ap1 (tp + 2*n)	/* n + 1 */
#define bp1 (tp + 3*n+1)	/* n + 1 */
#define xp (tp + 4*n + 2)	/* 2n + 2 */
#define scratch_out (tp + 6*n + 4)

      cy = mpn_add_n (am1, a0, a1, n);
      MPN_INCR_U (am1, n, cy);
      cy = mpn_sub_n (ap1, a0, a1, n);
      ap1[n] = mpn_add_1 (ap1, ap1, n, cy);

      cy = mpn_add_n (bm1, b0, b1, n);
      MPN_INCR_U (bm1, n, cy);
      cy = mpn_sub_n (bp1, b0, b1, n);
      bp1[n] = mpn_add_1 (bp1, bp1, n, cy);

      mul_2nm1_dc (rp, am1, bm1, n, scratch_out);
      mul_2np1_wrap (xp, ap1, bp1, n);

      /* xp = xm - xp mod (B^n + 1). Assumes normalized
	 representation. Puts high bit in hi. */
      if (UNLIKELY (xp[n]))
	hi = mpn_add_1 (xp, rp, n, 1);
      else
	{
	  cy = mpn_sub_n (xp, rp, xp, n);
	  hi = mpn_add_1 (xp, xp, n, cy);
	}
      /* Multiply by -B^n/2, using

	 -B^n/2 * (2 x1 + x0) = x1 - B^n/2 x0 (mod (B^n + 1))
      */

      cy = mpn_rshift (rp+n, xp, n, 1);
      if (hi != cy)
	rp[2*n-1] |= GMP_LIMB_HIGHBIT;
      if (hi < cy)
	/* Underflow */
	hi = mpn_add_1 (rp+n, rp+n, n, 1);
      else
	hi = 0;
      
      cy = mpn_sub_n (rp, rp, rp+n, n);
      cy = mpn_sub_1 (rp+n, rp+n, n, cy + hi);
      ASSERT (cy == hi);
#undef a0
#undef a1
#undef b0
#undef b1
#undef am1
#undef bm1
#undef ap1
#undef bp1
#undef xp
#undef scratch_out
    }
}

# if defined(__GNUC__) && defined(__i386__)
#define GET_CYCLE_COUNTER(hi, lo)		\
  __asm__("xorl %%eax,%%eax\n"			\
	  "movl %%ebx, %%edi\n"			\
	  "cpuid\n"				\
	  "rdtsc\n"				\
	  "movl %%edi, %%ebx\n"			\
	  : "=a" (lo), "=d" (hi)		\
	  : /* No inputs. */			\
	  : "%edi", "%ecx", "cc")

static inline double
cputime (void)
{
  int hi, lo;
  GET_CYCLE_COUNTER (hi, lo);
  return ldexp(hi, 32) + lo;
};

#define TIME_MIN_CYCLES 10000
#define TIME(t,func)							\
  do { long int __times, __t;						\
       double __t0, __tmp;						\
    __times = 1;							\
    {func;}								\
    do									\
      {									\
	__times <<= 1;							\
	__t0 = cputime ();						\
	for (__t = 0; __t < __times; __t++)				\
	  {func;}							\
	__tmp = cputime () - __t0;					\
      }									\
    while (__tmp < TIME_MIN_CYCLES);					\
    (t) = (double) __tmp / __times;					\
  } while (0)

static int
compare_double(const void *ap, const void *bp)
{
  double a = *(const double *) ap;
  double b = *(const double *) bp;
  if (a < b)
    return -1;
  else if (a > b)
    return 1;
  else
    return 0;
}
#endif

#define TEST_SIZE 512
#define MEDIAN_SIZE 5
int
main (int argc, char ** argv)
{
  mp_limb_t a[TEST_SIZE];
  mp_limb_t b[TEST_SIZE];
  mp_limb_t r[TEST_SIZE*2];
  mp_limb_t p[TEST_SIZE*2];
  mp_limb_t scratch[MUL_2NM1_DC_ITCH(TEST_SIZE)];
  
  mp_size_t n;
  mp_size_t old_dc_threshold;

  for (n = 2; n <= TEST_SIZE; n++)
    {
      mpn_random (a, n);
      mpn_random (b, n);
      mul_2nm1_wrap (r, a, b, n);

      if ((n & 1) == 0)
	{
	  mul_2nm1_x2 (p, a, b, n, scratch);
	  if (mpn_cmp (r, p, n) != 0)
	    {
	      gmp_printf ("Results differ for mul_2nm1_x2: n = %d\n"
			  "  a = %Nx\n"
			  "  b = %Nx\n"
			  "  r = %Nx\n"
			  "  p = %Nx\n",
			  n, a, n, b, n, r, n, p, n);
	      return EXIT_FAILURE;
	    }
	}
      
      if ((n & 3) == 0)
	{
	  mul_2nm1_x4 (p, a, b, n, scratch);
	  if (mpn_cmp (r, p, n) != 0)
	    {
	      gmp_printf ("Results differ for mul_2nm1_x4: n = %d\n"
			  "  a = %Nx\n"
			  "  b = %Nx\n"
			  "  r = %Nx\n"
			  "  p = %Nx\n",
			  n, a, n, b, n, r, n, p, n);
	      return EXIT_FAILURE;
	    }
	}
      mul_2nm1_dc (p, a, b, n, scratch);
      if (mpn_cmp (r, p, n) != 0)
	{
	  gmp_printf ("Results differ for mul_2nm1_dc: n = %d\n"
		      "  a = %Nx\n"
		      "  b = %Nx\n"
		      "  r = %Nx\n"
		      "  p = %Nx\n",
		      n, a, n, b, n, r, n, p, n);
	  return EXIT_FAILURE;
	}
    }
#ifdef GET_CYCLE_COUNTER
  old_dc_threshold = dc_threshold;

  printf("   n    mul  mullo mul_wrap x^2-1   x^4-1     dc1   dc (cycles, ratio)\n");
  for (n = 2; n <= TEST_SIZE; n += 2)
    {
      double elapsed[MEDIAN_SIZE];
      double t_mul, t_mullo, t_mul_wrap, t_mul_x2, t_mul_x4, t_mul_dc1, t_mul_dc;
      int i;
      double best;
      int zeros;
      
      mpn_random (a, n);
      mpn_random (b, n);

      for (i = 0; i < MEDIAN_SIZE; i++)
	TIME(elapsed[i], mpn_mul_n (r, a, b, n));
      qsort (elapsed, MEDIAN_SIZE, sizeof(double), compare_double);
      t_mul = elapsed[MEDIAN_SIZE/2];

      for (i = 0; i < MEDIAN_SIZE; i++)
	TIME(elapsed[i], mpn_mullow_n (r, a, b, n));
      qsort (elapsed, MEDIAN_SIZE, sizeof(double), compare_double);
      t_mullo = elapsed[MEDIAN_SIZE/2];

      for (i = 0; i < MEDIAN_SIZE; i++)
	TIME(elapsed[i], mul_2nm1_wrap (r, a, b, n));
      qsort (elapsed, MEDIAN_SIZE, sizeof(double), compare_double);
      t_mul_wrap = elapsed[MEDIAN_SIZE/2];

      for (i = 0; i < MEDIAN_SIZE; i++)
	TIME(elapsed[i], mul_2nm1_x2 (r, a, b, n, scratch));
      qsort (elapsed, MEDIAN_SIZE, sizeof(double), compare_double);
      t_mul_x2 = elapsed[MEDIAN_SIZE/2];
      
      if ((n & 2) == 0)
	{
	  for (i = 0; i < MEDIAN_SIZE; i++)
	    TIME(elapsed[i], mul_2nm1_x4 (r, a, b, n, scratch));
	  qsort (elapsed, MEDIAN_SIZE, sizeof(double), compare_double);
	  t_mul_x4 = elapsed[MEDIAN_SIZE/2];	  
	}
      else
	t_mul_x4 = 0.0;

      dc_threshold = n;

      for (i = 0; i < MEDIAN_SIZE; i++)
	TIME(elapsed[i], mul_2nm1_dc (r, a, b, n, scratch));
      qsort (elapsed, MEDIAN_SIZE, sizeof(double), compare_double);
      t_mul_dc1 = elapsed[MEDIAN_SIZE/2];

      dc_threshold = old_dc_threshold;
      
      for (i = 0; i < MEDIAN_SIZE; i++)
	TIME(elapsed[i], mul_2nm1_dc (r, a, b, n, scratch));
      qsort (elapsed, MEDIAN_SIZE, sizeof(double), compare_double);
      t_mul_dc = elapsed[MEDIAN_SIZE/2];

      best = t_mul_wrap; i = 0;
      if (t_mul_x2 < best)
	{ best = t_mul_x2; i = 1; }
      if (t_mul_x4 < best && (n & 2) == 0)
	{ best = t_mul_x4; i = 2; }
      if (t_mul_dc1 < best)
	{ best = t_mul_dc1; i = 3; }
      if (t_mul_dc < best)
	i = 4;

      count_trailing_zeros (zeros, n);
      printf("%4d %6.0f %6.3f %6.3f%c %6.3f%c %6.3f%c %6.3f%c %6.3f%c %d\n",
	     n, t_mul, t_mullo / t_mul,
	     t_mul_wrap / t_mul, i == 0 ? '#' : ' ',
	     t_mul_x2 / t_mul, i == 1 ? '#' : ' ',
	     t_mul_x4 / t_mul, i == 2 ? '#' : ' ',
	     t_mul_dc1 / t_mul, i == 3 ? '#' : ' ',
	     t_mul_dc / t_mul, i == 4 ? '#' : ' ',
	     zeros);
    }
#endif
  return EXIT_SUCCESS;
}
